<?php
/**
 * @file
 * Drupal Inspect module base class
 */

/**
 * Inspect base class
 *
 * @uses PHP:ini:precision
 * @static
 * @singleton
 * @category Drupal
 * @package Developer
 * @subpackage Inspect */
class InspectBs {
  /**
   * Prevents erratic instantiation of this entirely static class
   */
  private function  __construct() {}

  /**
   * Maximum sub var recursion depth.
   *
   * @type integer
   */
  const DEPTH_MAX = 10;

  /**
   * Default sub var recursion depth.
   *
   * @type integer
   */
  const DEPTH_DEFAULT = 10;

  /**
   * Minimum string truncation.
   *
   * @type integer
   */
  const TRUNCATE_MIN = 10000;

  /**
   * Default string truncation.
   *
   * @type integer
   */
  const TRUNCATE_DEFAULT = 100;

  /**
   * Base 62 uniqid() if session started, otherwise empty.
   *
   * @var string $_sessionNo
   */
  protected static $_sessionNo = '';

  /**
   * Request number in current session.
   *
   * @var string $_requestNo
   */
  protected static $_requestNo = 0;

  /**
   * Log number in current request.
   *
   * @var integer $_logNo
   */
  protected static $_logNo = 0;

  /**
   * Request time start in milliseconds.
   *
   * @var float $_requestTimeMilli
   */
  protected static $_requestTimeMilli = 0;

  /**
   * Default inspection options.
   * 
   *   Various options:
   *   - (array) category: logging category
   *   - (boolean) file_line: display file and line of call to inspect function/method
   *   - (array|NULL) filter: filter out that|those key name(s)
   *   - (boolean) one_lined: for creating one-lined formatting instead of multi-lined
   *
   *   String options:
   *   - (array) needles: list of char|string needles to be used in str_replace
   *   - (array) replacers: list of char|string replacers to be used in str_replace
   *   - (string) quote_begin: char|string used as initial quote
   *   - (string) quote_end: char|string used as final quote
   *
   *   Formatting options:
   *   - (array) delimiters: between object/array buckets, and for InspectTrc between frames and frame buckets
   *   - (string) indent: indentation of object/array buckets
   *   - (string) enclose_tag_get: html enclosure tag when getting
   *   - (string) enclose_tag_log: html enclosure tag when logging
   *   - (string) newline_get: newline character or tag when getting
   *   - (string) newline_log: newline character or tag when logging
   *
   * @var array $_options
   */
  protected static $_options = array(
    'category' => 'inspect',
    'file_line' => TRUE,
    'filter' => NULL,
    'one_lined' => FALSE,
    'needles' => array( "\n", "\r", "\t", '<', '>', '"', "'" ),
    'replacers' => array( '_NL_', '_CR_', '_TB_', '&#60;', '&#62;', '&#34;', '&#39;' ),
    'quote_begin' => '&#62;&#62;',
    'quote_end' => '&#60;&#60;',
    'delimiters' => array(
        "\n", // Before first bucket.
        "\n", // Before later bucket.
        "\n"  // Before last bucket.
    ),
    'indent' => '.  ',
    'enclose_tag_get' => 'pre',
    'enclose_tag_log' => 'pre',
    'newline_get' => "\n",
    'newline_log' => "\n",
  );

  /**
   * Options overriding the default, when truthy option one_lined.
   *
   * Apart form formatting, also makes default depth 1 and call file and line default off.
   *
   * @var array $_options_oneLined
   */
  protected static $_options_oneLined = array(
    'file_line' => FALSE,
    'delimiters' => array('', ', ', ''),
    'indent' => '',
    'enclose_tag_get' => '',
    'newline_log' => ' ',
    'newline_get' => ' ',
    'no_preface' => TRUE,
    'name' => 'none',
  );

  /**
   * Whether server OS is *nix (non-windows, really).
   *
   * @var integer $_os_nix
   */
  protected static $_os_nix = -1; // never checked

  /**
   * Whether the PHP multibyte extension exists and is loaded.
   *
   * @var integer $_mbString_exists
   */
  protected static $_mbString_exists = -1; // never checked

  /**
   * Document root of the current site, without trailing slash, and as regular expression (/^.../).
   *
   * If OS isnt *nix the value is an array: first bucket forward slashed, second bucket backslashed.
   *
   * @var string|array $_docRoot
   */
  protected static $_docRoot = ''; // never checked

  /**
   * Length of document root.
   *
   * @var integer $_docRootLength
   */
  protected static $_docRootLength; // never checked

  /**
   * Cache of the module permissions.
   *
   * @var array $_permissions
   */
  private static $_permissions = array(
    'inspect log' => FALSE,
    'inspect get' => FALSE,
  );

  /**
   * _inspect() returns empty if this isnt truthy.
   *
   * @var bool $_initialized
   */
  private static $_initialized = FALSE;

  /**
   * Sets/retrieves dedicated session number hash and request number in either session or cookie.
   *
   * If visitor starts without session, the value will be carried over from cookie to session when session is started,
   * whereas the opposite effectively wont happen (no carry-over from session to cookie).
   *
   * @return void
   */
  protected static function _initSessionNo() {
    if (self::$_sessionNo) {
      return;
    }
    if (drupal_session_started()) {
      $found = FALSE;
      $a =& $_SESSION;
      if (array_key_exists('module', $a)) {
        $a =& $a['module'];
        if (array_key_exists('inspect', $a)) {
          $a =& $a['inspect'];
          if (array_key_exists('sessionNo', $a)) {
            $found = TRUE;
            self::$_sessionNo = $a['sessionNo'];
            self::$_requestNo = (++$a['requestNo']);
          }
        }
        else {
          $a['inspect'] = array();
        }
      }
      else {
        $a['module'] = array('inspect' => array());
      }
      if (!$found) {
        $a =& $_SESSION['module']['inspect'];
        //  If cookie value exists from prior to session start: use that, and delete cookie.
        if (isset($_COOKIE['module_inspect_session'])
            && ($le = strlen($_COOKIE['module_inspect_session']) < 25) // Dont remember how large it may get, but that suffices.
            && preg_match('/^[a-zA-Z\d]+\:\d{1,5}$/', $_COOKIE['module_inspect_session'])
        ) {
          $c = explode(':', $_COOKIE['module_inspect_session']);
          self::$_sessionNo = $a['sessionNo'] = $c[0];
          self::$_requestNo = $a['requestNo'] = ((int)$c[1] + 1);
          setcookie (
              'module_inspect_session', '', time() - 3600,
              $GLOBALS['base_path'], $GLOBALS['cookie_domain'], $GLOBALS['is_https'], TRUE
          );
        }
        else {
          $comps = explode('.', uniqid('', TRUE));
          self::$_sessionNo = $a['sessionNo'] = self::baseConvert($comps[0], 16) . self::baseConvert($comps[1]);
          self::$_requestNo = $a['requestNo'] = 1;
        }
      }
      unset($a);
    }
    else { // Fall back on cookie.
      if (isset($_COOKIE['module_inspect_session'])
          && ($le = strlen($_COOKIE['module_inspect_session']) < 40) // Dont remember how large it may get, but that suffices.
          && preg_match('/^[a-zA-Z\d]+\:\d+$/', $_COOKIE['module_inspect_session'])
      ) {
        $a = explode(':', $_COOKIE['module_inspect_session']);
        self::$_sessionNo = $sess = $a[0];
        self::$_requestNo = $req = ((int)$a[1] + 1);
      }
      else {
        $comps = explode('.', uniqid('', TRUE));
        self::$_sessionNo = $sess = self::baseConvert($comps[0], 16) . self::baseConvert($comps[1]);
        self::$_requestNo = $req = 1;
      }
      setcookie (
          'module_inspect_session', $sess . ':' . $req, 0,
          $GLOBALS['base_path'], $GLOBALS['cookie_domain'], $GLOBALS['is_https'], TRUE
      );
    }
  }

  /**
   * @return array|bool
   *  - array: copy of self::$_permissions
   *  - FALSE if no user object, or not permitted to either log or get
   */
  protected static function _init() {
    if (!self::$_initialized) {
      if (!array_key_exists('user', $GLOBALS) || !is_object($GLOBALS['user']) || !property_exists($GLOBALS['user'], 'uid')) {
        return FALSE;
      }
      $log = $get = FALSE;
      if (user_access('inspect log')) {
        $log = self::$_permissions['inspect log'] = TRUE;
      }
      if (user_access('inspect get')
          //  ...And isnt admin user 1 (or is user 1 but not prohibited to do get).
          && ($GLOBALS['user']->uid != 1 || !variable_get('inspect_disallow_get_admin', FALSE))) {
        $get = self::$_permissions['inspect get'] = TRUE;
      }
      if (!$log && !$get) {
        return FALSE;
      }
      //  Check OS.
      $osNix = self::$_os_nix = strpos(PHP_OS, 'WIN') !== 0 ? 1 : 0;
      //  Find document root, for removal from absolute paths.
      self::$_docRootLength = strlen(
          $docRoot = substr(dirname(__FILE__), 0, -1 * (1 + strlen(drupal_get_path('module', 'inspect'))))
        );
      self::$_docRoot = $osNix ?
          ('/^' . preg_quote($docRoot, '/') . '/') :
          array(
            '/^' . preg_quote( str_replace('\\', '/', $docRoot), '/') . '/',  // all /
            '/^' . preg_quote( str_replace('/', '\\', $docRoot), '/') . '/'  // all \
          );
      //  Change file_line default?
      if (variable_get('inspect_fileline_omit', FALSE)) {
        static::$_options['file_line'] = FALSE;
      }
      //  Set/get session and request number.
      if (!self::$_sessionNo) {
        self::_initSessionNo();
      }
      //  Establish request time start.
      self::requestTimeMilli();

      self::$_initialized = TRUE;
    }
    return self::$_permissions; // Copy
  }

  /**
   * Checks permissions, secures that all inspection options exist and are in accord with permissions.
   *
   * Any method calling InspectBs::_inspect() must guarantee to call this method at least once before calling _inspect().
   *
   * If the calling method wants to call _inspect() more than once it must guarantee to re-use the return value of this method
   * for the later calls to _inspect().
   *
   * Sets/updates $_SESSION['module']['inspect']['sessionNo'] if session started (only on first call, initializing).
   *
   * @see InspectBs::$_options
   * @param string $target
   *   - allowed values: get|log
   * @param mixed $options
   *   - see InspectBs::$_options for supported options (plus depth and truncate)
   *   - default: NULL
   *   - array|object: list of options
   *   - string: interprets to message
   *   - integer: interprets to suggested depth
   *   - other type: ignored
   * @return array|bool|NULL
   *   - FALSE if the user isnt allowed to get|log inspections
   *   - NULL if arg $target isnt get or log
   */
  protected static function _checkPermissionsAndOptions($target, $options = NULL) {
    if (!self::$_initialized && !self::_init()) {
      return FALSE;
    }
    //  Check general permissions.
    $hideScalars = FALSE;
    $hidePaths = FALSE;
    if ($target === 'get') {
      if (!self::$_permissions['inspect get']) {
        return FALSE;
      }
      $hidePaths = TRUE; // we always hide paths when getting
    }
    elseif ($target === 'log') {
      if (!self::$_permissions['inspect log']) {
        return FALSE;
      }
    }
    else {
      return NULL;
    }
    //  Check/load options.
    $isList = $isArr = FALSE;
    $depthOpt = static::DEPTH_DEFAULT;
    $truncOpt = static::TRUNCATE_DEFAULT;
    $optsDef =& static::$_options;
    switch (gettype($options)) {
      case 'NULL':
        $opts = $optsDef; // copy
        break;
      case 'array':
        if (count($options)) {
          $isList = $isArr = TRUE;
          $opts = $options;
        }
        else {
          $opts = $optsDef; // copy
        }
        break;
      case 'object':
        $isList = TRUE;
        $opts = get_object_vars($options);
        break;
      case 'integer':
        $depthOpt = $options;
        $opts = $optsDef; // copy
        break;
      case 'string':
        $opts = $optsDef; // copy
        $opts['message'] = $options;
        break;
      default:
        //  Ignore argument.
        $opts = $optsDef; // copy
    }
    //  If list as argument, check validity of the options.
    if ($isList) {
      //  Make sure string needles and replacers arent falsy.
      if (isset($opts['needles']) && !$opts['needles']) {
        unset($opts['needles']);
      }
      if (isset($opts['replacers']) && !$opts['replacers']) {
        unset($opts['replacers']);
      }
      if (empty($opts['one_lined'])) {
        $oneLined = FALSE;
      }
      else if ($target === 'get') {
        $oneLined = TRUE;
        $depthOpt = 1;
      }
      else {
        unset($opts['one_lined']);
      }
      //  Read depth, hide scalars, truncate, filter, hide paths, if they exist.
      if (isset($opts['depth'])) {
        $depthOpt = (int)$opts['depth'];
      }
      if (!$hideScalars && isset($opts['hide_scalars']) && $opts['hide_scalars']) {
        $hideScalars = TRUE;
      }
      if (isset($opts['truncate'])) {
        if (($v = (int)$opts['truncate'])) {
          if ($v < 0) {
            $v = 0;
          }
          else if ($v > static::TRUNCATE_MIN) { // If suggested truncate is more than overall min.
            $v = static::TRUNCATE_MIN;
          }
        }
        $truncOpt = $v;
      }
      if (isset($opts['filter'])) {
        if (is_array($v = $opts['filter'])) {
          if (!count($v)) {
            unset($opts['filter']);
          }
        }
        elseif (strlen('' . $v)) {
          $opts['filter'] = array(0 => '' . $v);
        }
      }
      //  We set the bucket hide_paths no matter what, and it can only override in negative (safer) manner.
      if (!$hidePaths && isset($opts['hide_paths']) && $opts['hide_paths']) {
        $hidePaths = TRUE;
      }
      //  If array, some options may also be set as value (and thus numeric key) ~ interpreted as truthy boolean.
      if ($isArr) {
        $a = $opts; // Iterate copy, to allow modification.
        foreach ($a as $k => $v) {
          if (is_string($v) && ('' . (int)$k) == ('' . $k)) { // ~ Integer key, don't rely on type of PHP array numeric keys.
            switch ($v) {
              case 'hide_scalars':
                $opts['hide_scalars'] = $hideScalars = TRUE;
                unset($opts[$k]);
                break;
              case 'hide_paths':
                $opts['hide_paths'] = $hidePaths = TRUE;
                unset($opts[$k]);
                break;
              case 'one_lined':
                if ($target === 'get') {
                  $opts['one_lined'] = $oneLined = TRUE;
                }
                unset($opts[$k]);
                break;
              case 'no_folding':
              case 'no_preface':
                $opts[$v] = TRUE;
                unset($opts[$k]);
                break;
            }
          }
        }
      }
      //  load defaults, and make arg options override
      $opts = !$oneLined ? array_merge($optsDef, $opts) : array_merge($optsDef, static::$_options_oneLined, $opts);
    }
    //  Check depth.
    if ($depthOpt < 0) {
      $depthOpt = 0;
    }
    elseif ($depthOpt && $depthOpt > static::DEPTH_MAX) {
      $depthOpt = $depth_max;
    }
    //  Load resolved values of depth and truncate into the options array.
    $opts['depth'] = $depthOpt;
    $opts['hide_scalars'] = $hideScalars;
    $opts['truncate'] = $truncOpt;
    $opts['hide_paths'] = $hidePaths;
    //  Escape name and message options
    if(!empty($opts['name'])) {
      $opts['name'] = htmlspecialchars(
          $opts['name'],
          ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
          'UTF-8',
          FALSE // No double encoding
      );
    }
    else {
      $opts['name'] = '';
    }
    if(!empty($opts['message'])) {
      $opts['message'] = htmlspecialchars(
          $opts['message'],
          ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
          'UTF-8',
          FALSE // No double encoding
      );
    }
    //  Flag whether getting or logging.
    //$opts['target'] = $target; - not used currently
    return $opts;
  }

  /**
   * Resolve options, make sure it's an array.
   *
   * @param mixed $options
   * @param string $integer
   *  - default: depth
   *  - what to interpretate $options as, if $options is an integer
   * @param string $string
   *  - default: message
   *  - what to interpretate $options as, if $options is a string
   * @return array
   */
  protected static function _resolveOptions($options, $integer = 'depth', $string = 'message') {
    switch (gettype($options)) {
      case 'array':
        $opts = $options;
        break;
      case 'object':
        return get_object_vars($options);
      case 'integer':
        return array($integer => $options);
      case 'string':
        return array($string => $options);
      default:
        return array();
    }
    //  If array, some options may also be set as value (and thus numeric key) ~ interpreted as truthy boolean.
    $a = $opts; // Iterate copy, to allow modification.
    foreach ($a as $k => $v) {
      if (is_string($v) && ('' . (int)$k) == ('' . $k)) { // ~ Integer key, don't rely on type of PHP array numeric keys.
        switch ($v) {
          case 'hide_scalars':
          case 'hide_paths':
          case 'one_lined':
          case 'no_folding':
          case 'no_preface':
            //  Clean up, make value key.
            unset($opts[$k]);
            $opts[$v] = TRUE;
            break;
        }
      }
    }
    return $opts;
  }

  /**
   * The actual variable analyser - do not call it with an options arg that hasnt been checked by _checkPermissionsAndOptions()
   *
   * See InspectVar::get() for specifics of arg $options.
   *
   *   String truncation:
   *   - truncate option less than one (hide strings) doesnt apply to strings that solely consist of digits (stringed integers)
   *   - a string bucket of an object/array keyed pass or password (lowercase) will always be hidden (no matter it's content)
   *
   *   Removes (what looks like) paths when getting (not logging), relative as well as absolute path:
   *   - if *nix server OS: only strings having no space(s), unix style absolute paths, and only looks for forward slash
   *   - if windows server OS: windows style absolute paths, and looks for forward and backward slashes
   *   - path removal takes place before escaping and truncating (except for overly long strings)
   *
   *   Reports non-integer numbers:
   *   - double as float
   *   - NaN as NaN and infinite as infinite (unless hiding scalars, in such case integer|float)
   *
   *   Reports string lengths and path removal (path reduced to filename):
   *   - un-truncated string, and when hiding scalar values: (multibyte length|ascii length)
   *   - truncated string: (multibyte length|ascii length|truncation length)
   *   - if path removed: there will be a fourth flag (exclamation mark), if no truncation truncation length will be hyphen
   *
   *   Recursive:
   *   - always stops at 10th (static::DEPTH_MAX) recursion or less (~ the depth option)
   *   - when buckets of an object/array is analyzed, buckets of buckets will be analyzed to one lesser depth
   *   - detects if a bucket of an object references the object itself (marked *RECURSION*)
   *   - detects GLOBALS array's self-referencing GLOBALS bucket (marked *RECURSION*)
   *   - doesnt check identity for arrays (not possible in PHP), nor 'same-ness' (too heavy, and prone to err)
   *
   * @see InspectVar::get()
   * @param mixed $var
   * @param array &$options
   *   - by reference
   * @param integer $_curDepth
   *   - default: zero
   *   - recursion counter, do not call it with anything but zero
   * @return string
   */
  protected static function _inspect($var, &$options, $_curDepth = 0) {
    if (!self::$_initialized) {
      return '';
    }
    $maxDepth = $options['depth'];
    //  Recursion limit check.
    if ($_curDepth
        && $_curDepth > $maxDepth) {
      return '_INSPECT_DEPTH_LIMIT_:' . $maxDepth;
    }
    //  Object or array.
    $tArr = FALSE;
    if (is_object($var) || ($tArr = is_array($var))) {
      $s = $tArr ? 'array' : ($tObj = get_class($var));
      if(!$_curDepth && !strlen($options['name'])) {
        $options['name'] = $s;
      }
      $s = '(' . $s . ':';
      //  If at max depth: simply get length of the container.
      if ($_curDepth == $maxDepth) {
        if ($tArr) {
          $s .= count($var) . ')';
        }
        else { // stdClass cannot be counted, simply iterate
          $le = 0;
          foreach ($var as $sub) {
            ++$le;
          }
          $s .= $le . ')';
        }
        return $s;
      }
      //  Dive into container buckets.
      else {
        $delimFirst = $options['delimiters'][0]
            . str_repeat($indent = $options['indent'], $_curDepth + 1);
        $delimLater = $options['delimiters'][1]
            . str_repeat($indent, $_curDepth + 1);
        $le = 0;
        $sSub = '';
        //  Different iterations for object vs array, because:
        //  - array ust iterate by explicit reference,
        //    whereas object may fail fatally doing that (An iterator cannot be used with foreach by reference)
        //  - they must have different self reference checks
        //  - array must be checked for (all) numeric indices
        if ($tArr) {
          $numArr = TRUE; // if all numeric indices, use [] instead of {}
          foreach ($var as $key => &$sub) { // by explicit reference
            $delim = (++$le) > 1 ? $delimLater : $delimFirst;
            //  GLOBALS check, skip self reference.
            if ($key === 'GLOBALS' && is_array($sub) && array_key_exists('GLOBALS', $sub)) {
              $sSub .= $delim . 'GLOBALS: (array) *RECURSION*';
              $numArr = FALSE;
              continue;
            }
            if ($numArr && $key !== $le - 1) { // numeric array check
              $numArr = FALSE;
            }
            if ($options['filter'] && in_array($key, $options['filter'], TRUE)) {
              $sSub .= $delim . $key . ': F';
            }
            else {
              $options['key'] = $key; // pass key, for password check
              $sSub .= $delim . $key . ': ' . self::_inspect($sub, $options, $_curDepth + 1); // recursion
            }
          }
          return $s . $le . ') '
              . ($numArr ? '[' : '{')
              . $sSub . $options['delimiters'][2] . str_repeat($indent, $_curDepth)
              . ($numArr ? ']' : '}');
        }
        else { // object
          foreach ($var as $key => $sub) { // not &$sub, iterating object is implicitly by reference
            $delim = (++$le) > 1 ? $delimLater : $delimFirst;
            if ($sub === $var) { // check for identity (self reference)
              $sSub .= $delim . $key . ': (' . $tObj . ') *RECURSION*';
              continue;
            }
            if ($options['filter'] && in_array($key, $options['filter'], TRUE)) {
              $sSub .= $delim . $key . ': F';
            }
            else {
              $options['key'] = $key; // pass key, for password check
              $sSub .= $delim . $key . ': ' . self::_inspect($sub, $options, $_curDepth + 1); // recursion
            }
          }
          return $s . $le . ') {'
              . $sSub . $options['delimiters'][2] . str_repeat($indent, $_curDepth)
              . '}';
        }
      }
    }
    //  Scalars, NULL and resource.
    switch (($t = gettype($var))) {
      case 'boolean':
        return '(' . $t . ')' . (!$options['hide_scalars'] ? ($var ? ' TRUE' : ' FALSE') : '');
      case 'integer':
      case 'double':
      case 'float':
        if (!$options['hide_scalars'] && !is_finite($var)) {
          return '(' . (is_nan($var) ? 'NaN' : 'infinite') . ')';
        }
        return '(' . ($t == 'double' ? 'float' : $t) . ')'
            . ( $options['hide_scalars'] ? '' :
                (!$var ? ' 0' : (' ' . self::numberToString($var)) ) );
      case 'string':
        //  (string:n0|n1|n2): n0 ~ multibyte length, n1 ~ ascii length, n2 only occurs if truncation.
        $s = '(string:';
        //  Get multibyte length and ascii length, and check for invalid UTF-8.
        $leRaw = strlen($var);
        if (!drupal_validate_utf8($var)) {
          return $s . '?|' . $leRaw . '|0) *INVALID_UTF8*';
        }
        $leMb = $leRaw ? self::_mb_strlen($var) : 0;
        $s .= !$leRaw ? '0|0' : ($leMb . '|' . $leRaw);
        //  If empty or hide scalar values.
        if ($options['hide_scalars'] || !$leRaw) {
          return $s . ')'
              . ( $options['hide_scalars'] ? '' : // no quotes if hide value
                  (' ' . $options['quote_begin'] . $options['quote_end']) );
        }
        //  If very long string, truncate to absolute max before working on the string internals.
        $truncedTo = 0;
        if ($leMb > ($trunc = $options['truncate']) && $leMb > ($trunc_min = static::TRUNCATE_MIN)) {
          $var = self::_mb_truncate($var, $trunc_min);
          $truncedTo = $leMb = $trunc_min;
        }
        //  If truncate is zero.
        if (($pw = ($_curDepth > 0 && ($options['key'] === 'pass' || $options['key'] === 'password'))) || $trunc < 1) {
          //  Look for stringed integer (consists solely of digits), but only if not truncated.
          if (!$pw && !$truncedTo && preg_match('/^\d+$/', $var)) {
            return $s . ') ' . $options['quote_begin'] . $var . $options['quote_end'];
          }
          return $s . '|0)'; // |0 ~ flag truncation
        }
        //  Remove document root (always) and/or path.
        //  Do it before escaping (needles may contain dot or slash).
        $pathRem = FALSE;
        if ($leRaw > 3) {
          if ($options['hide_paths']) { //  This algo also removes document root.
            if (self::$_os_nix) { // Non-windows.
              if (strpos($var, ' ') === FALSE // If no space in string.
                  && ($posSlash = strpos($var, '/') ) !== FALSE) { // And has forward slash.
                if (!$posSlash // Slash at pos zero ~ absolute path.
                    || ( $var{0} === '.'
                        && (strpos($var, '../') === 0 || strpos($var, './') === 0 ) ) // Relative.
                ) {
                  $pathRem = TRUE;
                }
              }
            }
            else { // Windows.
              if ( strpos($var, '/') || strpos($var, '\\') ) { // has slash, windows path cant have slash at pos zero
                if (($char = $var{0}) === '.') { // check for relative path
                  if (strpos($var, '../') === 0 || strpos($var, './') === 0
                        || strpos($var, '..\\') === 0 || strpos($var, '.\\') === 0) {
                    $pathRem = TRUE;
                  }
                }
                //  Absolute path (ascii letter + : + slash).
                //elseif (preg_match('/^[A-Za-z]\:[\/\x5C]/', $var)) {
                elseif ( ( ( ($char = ord($char)) >= 65 && $char <= 90) || ($char >= 97 && $char <= 122) )
                    && $var{1} === ':'
                    && (($char = $var{2}) === '/' || $char === '\\')
                ) {
                  $pathRem = TRUE;
                }
              }
            }
            if ($pathRem) {
              $var = basename($var);
            }
          }
          elseif ($leRaw > self::$_docRootLength) {
            if (self::$_os_nix) { // non-windows
              if ($var{0} === '/' && strlen($var = preg_replace(self::$_docRoot, '', $var)) < $leRaw) {
                $pathRem = TRUE;
                $truncedTo = $leMb = self::_mb_strlen($var);
                $var = 'document_root' . $var;
              }
            }
            else {
              if (strlen($var = preg_replace(self::$_docRoot, '', $var)) < $leRaw) {
                $pathRem = TRUE;
                $truncedTo = $leMb = self::_mb_strlen($var);
                $var = 'document_root' . $var;
              }
            }
          }
        }
        //  Escape string.
        $var = htmlspecialchars(
            str_replace($options['needles'], $options['replacers'], $var),
            ENT_QUOTES, // PHP 5.4: ENT_QUOTES | ENT_SUBSTITUTE
            'UTF-8',
            FALSE // No double encoding
        );
        //  Truncate string.
        if ($leMb > $trunc) {
          $var = self::_mb_truncate($var, $trunc);
          $truncedTo = $trunc;
        }
        return $s
            . (!$truncedTo ? (!$pathRem ? '' : ('|-|!') ) :
                ('|' . $truncedTo . (!$pathRem ? '' : ('|!') ) ) )
            . ') '
            . $options['quote_begin'] . $var . $options['quote_end'];
      case 'resource':
        return '(resource)' . (!$options['hide_scalars'] ? (' ' . get_resource_type($var) ) : '');
      default: // NULL and unknown type
        return '(' . $t . ')';
    }
  }

  /**
   * Get file and line of outmost call to an inspect function or public method.
   *
   * @param array|NULL $trace
   *  - slightly cheaper when given debug_backtrace() arg
   * @param boolean $hidePaths
   *  - default: TRUE
   * @return string
   */
  protected static function _fileLine($trace = NULL, $hidePaths = TRUE) {
    if ($trace) {
      if(($le = count($trace)) > 3) {
        $le = 3;
      }
    }
    else if(($le = count($trace = debug_backtrace())) > 4) {
      $le = 4;
    }
    //  finds last frame having an Inspect class or inspect function
    $iFrm = -1;
    for ($i = 0; $i < $le; $i++) {
      $nspct = FALSE;
      if (isset($trace[$i]['class']) && strlen($s = $trace[$i]['class'])) {
        $nspct = (strpos($s, 'Inspect') === 0);
      }
      else {
        $nspct = (isset($trace[$i]['function']) && strlen($s = $trace[$i]['function'])
              && ($s == 'inspect' || strpos($s, 'inspect_') === 0) );
      }
      if ($nspct) {
        $iFrm = $i;
      }
      else if ($iFrm > -1) { // Break a first non-inspect method/function after an inspect method/function was found.
        break;
      }
    }
    return '@'
        . ($iFrm > -1 ?
            (
                (isset($trace[$iFrm]['file']) && strlen($s = trim($trace[$iFrm]['file'])) ?
                  ($hidePaths ? basename($s) : ('document_root' . preg_replace(self::$_docRoot, '', $s)) ) :
                  'unknown'
                )
                . ':'
                . (isset($trace[$iFrm]['line']) ? trim($trace[$iFrm]['line']) : '?')
            ) : 'unknown');
  }


  /**
   * Resolve constants.
   *
   * @param string|object|falsy $cat_or_instance_or_classname
   *   - default: 'user' (~ the user category of defined constants)
   *   - falsy: all defined constants, listed by category
   *   - 'class SomeClassName': class constants of class SomeClassName
   *   - other strings: that category of defined constants
   *   - object: class constants of the object's class
   * @param bool $keySort
   *   - default: falsy (~ do not sort constants, by name)
   * @return array|NULL
   *   - array, buckets: array to be inspected, name
   *   - FALSE on error
   */
  protected static function _constants($cat_or_instance_or_classname = 'user', $keySort = FALSE) {
    $cat = $cat_or_instance_or_classname;
    $nm = 'constants';
    if (!$cat) {
      $a = get_defined_constants(TRUE);
    }
    elseif (is_string($cat)) {
      $cat = preg_replace('/[^a-zA-Z\d_\-\ ]/', '', $cat);
      if (substr($cat, 0, 6) === 'class ') {
        if (($c = substr($cat, 6)) && class_exists($c)) {
          $o = new ReflectionClass($c);
          $a = $o->getConstants();
          $nm = 'class ' . $c . ' ' . $nm;
        }
        else {
          return FALSE;
        }
      }
      else {
        $a = get_defined_constants($cat);
        $a = array_key_exists($cat, $a) ? $a[$cat] : array();
        $nm = $cat . ' ' . $nm;
      }
    }
    elseif (is_object($cat)) {
      $c = get_class($cat);
      $o = new ReflectionClass($c);
      $a = $o->getConstants();
      $nm = 'class ' . $c . ' ' . $nm;
    }
    else {
      return FALSE;
    }
    if ($keySort) {
      ksort($a);
    }
    return array($a, $nm);
  }

  /**
   * Get request time start - in milliseconds - more accurately than the REQUEST_TIME constant.
   *
   * The value is usually slightly higher than REQUEST_TIME, apparantly the latter is a floor'ed value.
   *
   * Uses either $_SERVER['REQUEST_TIME_FLOAT'] (PHP>=5.4), request header X-Request-Received-Processed, or REQUEST_TIME.
   *
   * To enable use of the custom request header (Apache only), add the following to document root .htaccess:
   * @code
   * <IfModule mod_headers.c>
   *   RequestHeader set X-Request-Received-Processing "%t %D"
   * </IfModule>
   * @endcode
   * 
   * @return float
   */
  public static function requestTimeMilli() {
    $rtm =& self::$_requestTimeMilli;
    if (!$rtm) {
      $rtm = ($rts = REQUEST_TIME) * 1000;
      if (array_key_exists('REQUEST_TIME_FLOAT', $_SERVER) && ($v = (int) $_SERVER['REQUEST_TIME_FLOAT'])) {
        $rtm = round($v * 1000, 3);
      }
      else if (($a = apache_request_headers()) && array_key_exists('X-Request-Received-Processing', $a)
          && strlen($v = $a['X-Request-Received-Processing']) < 31 && preg_match('/^t\=\d{16,17}\ D\=\d{1,9}/', $v)) {
        $a = explode(' ', $v);
        $proc = (float) substr($a[1], 2);
        //  If received time can be reduced to positive integer, and that is close to REQUEST_TIME.
        if ( ($iRec = (int) round(($flRec = (float) substr($a[0], 2)) / 1000000)) > 0 && $iRec <= $rts + 1 && $iRec >= $rts - 1) {
          $rtm = round(($flRec + $proc) / 1000, 3);
        }
      }
    }
    return $rtm;
  }

  /**
   * Log to database (currently watchdog).
   *
   * No permissions check for this method - everybody is allowed to log a message.
   *
   * @param string $message
   *   - default: empty string
   * @param string $category
   *   - default: inspect
   * @param integer $severity
   *   - default: WATCHDOG_DEBUG
   * @return bool
   */
  public static function log_db($message = '', $category = 'inspect', $severity = WATCHDOG_DEBUG) {
    watchdog(
        $category,
        $message,
        NULL,
        $severity > -1 ? $severity : WATCHDOG_DEBUG
    );
    return TRUE;
  }
  //  Utility methods.
  /**
   * Becomes the value of php:ini:precision.
   *
   * @var integer $_precision
   */
  protected static $_precision = 0;

  /**
   * Convert number to string avoiding E-notation for numbers outside system precision range.
   *
   * @param float|integer $num
   * @return string
   */
  public static function numberToString($num) {
    if (!($prec = self::$_precision)) {
      $prec = self::$_precision = ini_get('precision');
    }
    return ($num < pow(10, $prec) && $num > pow(10, -$prec)) ?  // if within system precision, just string it
        ('' . $num) : number_format($num, 0, '.', '');
  }

  /**
   * Convert number or hex string to any base through 2 and 62  (~ larger than 16).
   *
   * Base 62 produces up to 35% shorter string than base 16.
   *
   *  Like native base_convert() precision fails for numbers (somewhere) larger than (tested on windows box with ini precision 14):
   *  - base 10: 9007199254740991 (> 15 digits)
   *  - base 16: 1fffffffffffff (> 14 digits)
   *
   * @param integer|float|string $num
   * @param integer $fromBase
   *  - default: 10
   *  - only supports 10 and 16 (if other, returns arg $num unaltered)
   * @param integer $toBase
   *  - default (and max): 62
   * @return string
   */
  public static function baseConvert($num, $fromBase = 10, $toBase = 62) {
    $table = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';
    $zero = $table{0};
    $remainder = $num;
    if($num < 1) {
      if(!$num) {
        return '' . $zero; // zero
      }
      $remainder *= -1;
    }
    if($toBase == $fromBase) {
      return $remainder;
    }
    switch ($fromBase) {
      case 10:
        break;
      case 16:
        $remainder = base_convert($remainder, 16, 10);
        break;
      default:
        return $remainder;
    }
    if($toBase > 62) {
      $toBase = 62;
    }
    else if($toBase < 2) {
      $toBase = 2;
    }
    $res = '';
    //  Find power ~ required number of digits.
    $pow = 0;
    $divisor = 1;
    $limit = 100; // probably way too high, havent seen more than 10 iterations (1fffffffffffff, with precision 14)
    while ($remainder >= ($test = pow($toBase, ++$pow)) && (--$limit)) {
      $divisor = $test;
      --$limit;
    }
    --$pow;
    //  Do modulo procedure.
    $limit = 100; // probably way too high, havent seen more than 14 iterations (1fffffffffffff, with precision 14)
    while ($remainder > 0 && (--$limit)) {
      if ($pow) {
        if ($divisor == $remainder) {
          return $res . $table{1} . str_repeat($zero, $pow);
        }
        else if ($divisor > $remainder) {
          $res .= $zero;
        }
        else {
          if ($remainder > $divisor) {
            $res .= $table{ floor($remainder / $divisor) };
            $remainder = fmod($remainder, $divisor);
          }
          else {
            $res .= $table{ $pow };
            $remainder = fmod($remainder, $divisor);
          }
          if($remainder < 1 && !floor($remainder)) {
            return $res . str_repeat($zero, $pow);
          }
        }
      }
      else {
        return $res . $table{ $remainder };
      }
      $divisor = pow($toBase, --$pow);
    }
    return $res;
  }

  /**
   * Multibyte-safe string length counter which doesnt require the mb_string PHP extension.
   *
   * @param string $str
   * @return integer
   */
  protected static function _mb_strlen($str) {
    static $mb = -1;
    if (!($le = strlen($str))) {
      return 0;
    }
    if ($mb < 0) {
      if (($mb = self::$_mbString_exists) < 0) {
        $mb = self::$_mbString_exists = extension_loaded('mbstring') ? 1 : 0;
      }
    }
    if ($mb) {
      return mb_strlen($str);
    }
    $n = 0;
    $leading = FALSE;
    for ($i = 0; $i < $le; $i++) {
      if (($ord = ord($str{$i})) < 128) {
        ++$n;
        $leading = FALSE;
      }
      else if ($ord < 192) { // continuation char
        $leading = FALSE;
      }
      else { // leading char
        if (!$leading) { // a sequence of leadings only counts as a single
          ++$n;
        }
        $leading = TRUE;
      }
    }
    return $n;
  }

  /**
   * Multibyte-safe string truncator which doesnt require the mb_string PHP extension.
   * 
   * @param string $str
   * @param integer $length
   *  - default: zero
   * @return integer
   */
  protected static function _mb_truncate($str, $length = 0) {
    static $mb = -1;
    if ($length < 1 || !($le = strlen($str))) {
      return '';
    }
    if ($mb < 0) {
      if (($mb = self::$_mbString_exists) < 0) {
        $mb = self::$_mbString_exists = extension_loaded('mbstring') ? 1 : 0;
      }
    }
    if ($mb) {
      return mb_substr($str, 0, $length);
    }
    $n = 0;
    $leading = FALSE;
    for ($i = 0; $i < $le; $i++) {
      if (($ord = ord($str{$i})) < 128) {
        if ((++$n) > $length) {
          return substr($str, 0, $i);
        }
        $leading = FALSE;
      }
      else if ($ord < 192) { // continuation char
        $leading = FALSE;
      }
      else { // leading char
        if (!$leading) { // a sequence of leadings only counts as a single
          if ((++$n) > $length) {
            return substr($str, 0, $i);
          }
        }
        $leading = TRUE;
      }
    }
    return $str;
  }
}