<?php
/**
 * @file
 * Drupal Inspect module administration layer
 */

/**
 * Implements drupal_get_form().
 *
 * Defines configuration form fields.
 * 
 * @param array $form
 * @param array &$form_state
 * @return array
 *   - the return value of system_settings_form()
 */
function inspect_admin_form($form, &$form_state) {
  drupal_add_css(
      drupal_get_path('module', 'inspect') . '/inspect.admin.css',
      array('type' => 'file', 'group' => CSS_DEFAULT, 'preprocess' => false, 'every_page' => false)
  );
  $form['inspect_disallow_get_admin'] = array(
    '#type' => 'checkbox',
    '#title' => t('Prohibit the primary admin user from getting inspections of variables and traces'),
    '#description' => nl2br(t(
'Overrides the &#39;Get outputs of variable/trace inspections&#39; permission for user 1.
Does not prohibit user 1 from <em>logging</em> inspections.

See also <a href="@inspect_perms">Permissions</a> and <a href="@inspect_help">Help</a>.

Admin user #1 - the omnipotent user created when setting up a Drupal site - is not restricted by any permissions set
(revoked, disabled) in the permissions dialogue. The means for modules of checking permissioms - <a href="@api_user_access">user_access()</a>
- simply returns TRUE for user #1, unconditionally.
Checking this option amends that - however an even better way of handling this problem is never to use the admin user 1 account
- except when updating modules and core.',
        array(
          '@inspect_perms' => '/admin/people/permissions#module-inspect',
          '@inspect_help' => '/admin/help/inspect',
          '@api_user_access' => 'http://api.drupal.org/api/drupal/modules--user--user.module/function/user_access/7',
        )
    )),
    '#default_value' => variable_get('inspect_disallow_get_admin', FALSE),
    //'#access' => FALSE,
    '#required' => FALSE,
  );
  $form['inspect_fileline_omit'] = array(
    '#type' => 'checkbox',
    '#title' => t('Make backend inspections default to omit displaying file and line of inspection function/method call'),
    '#default_value' => variable_get('inspect_fileline_omit', FALSE),
    //'#access' => FALSE,
    '#required' => FALSE,
  );
  $form['inspect_output_nofolding'] = array(
    '#type' => 'checkbox',
    '#title' => t('Turn off variable dump folding (expandable/collapsible outputs)'),
    '#description' => nl2br(t('May also be turned off intermittently by calling <strong>inspect.toggleFolding()</strong> in the browser console.
And per inspection, by using the no_folding option (the $options argument).
    ')),
    '#default_value' => variable_get('inspect_output_nofolding', FALSE),
    //'#access' => FALSE,
    '#required' => FALSE,
  );
  return system_settings_form($form);
}

/**
 * Creates help page.
 *
 * Implements hook_help().
 * @param string $path
 * @return string
 */
function _inspect_help($path) {
  if ($path != 'admin/help#inspect') {
    return;
  }
  $module = 'inspect';
  $headline_tag = 'h3';
  $paragraph_start = 'p class="admin-help-module-' . $module . '-paragraph"';
  $module_path = drupal_get_path('module', $module);
  $primary_function = $module;

  drupal_add_css(
      $module_path . '/inspect.admin.css',
      array('type' => 'file', 'group' => CSS_DEFAULT, 'preprocess' => false, 'every_page' => false)
  );

  $out = '<div><div class="admin-help-module-' . $module . '-pane-first">';

  $out .= '
<' . $headline_tag . '>' . t('Inspect variables, constants, error- and back traces') . '</' . $headline_tag . '>
<h4>' . t('Backend as well as frontend') . '</h4>
<' . $paragraph_start . '>'
  . t(
'Please read the <a href="#readme">Readme</a> - particular the part on <a href="#permissions_and_safety">Permissions and safety</a>.'
  )
  . '</p><' . $paragraph_start . '>'
  . t(
'For full documentation of backend and frontend functions and methods - please check <a href="@doc_backend" target="_blank">phpdocs</a> and <a href="@doc_frontend" target="_blank">jsdocs</a>.',
    array(
      '@doc_backend' => 'http://www.simple-complex.net/docs/drupal/modules/inspect/php/inspect_8module.html',
      '@doc_frontend' => 'http://www.simple-complex.net/docs/drupal/modules/inspect/js/symbols/Inspect.html',
    )
  )
  . '</p><' . $paragraph_start . '>'
  . t(
'The options and workings of the module\'s primary function - inspect() - is also documented here'
  )
  . ':</p>';

  //  Read primary function's documentation.
  if (strlen($read_source = @file_get_contents($module_path . '/' . $module . '.module'))) {
    $pos_end = strpos($read_source, "\nfunction " . $primary_function . '(');
    $s_start = substr($read_source, 0, $pos_end);
    $s_start = substr($s_start, strrpos($s_start, '/**'));
    $s_end = substr($read_source, $pos_end);
    $pos_end = strpos($s_end, ')');
    $s = str_replace("\r", '', $s_start . substr($s_end, 0, $pos_end + 1) );
    $s = preg_replace('/\n([^\n]+)$/', "\n" . '<strong>$1 {...}</strong>',
        preg_replace('/^([^\n]+)\n/', '<strong>$1</strong>' . "\n",
        str_replace("\n- ", "\n&bull; ", // hyphen to bullet
        str_replace("\n/", "\n", // comment end
        preg_replace('/\n \*[ \t]*/', "\n", // comment line *
        str_replace('/**', '', // comment start
      $s))))));
    $out .= '<' . $paragraph_start . '>'
        . nl2br(
            preg_replace('/\n\n/', '</p><' . $paragraph_start . '>', $s)
        ) . '</p>';
  }

  $out .= '</div><div class="admin-help-module-' . $module . '-pane-second">';
  //  Parse README.txt to html.
  $out .= '<a name="readme"></a><h3>Readme</h3>';
  $readme = @file_get_contents(
      substr(dirname(__FILE__), 0, -1 * (1 + strlen($module_path)))
      . '/' . drupal_get_path('module', 'inspect')
      . '/README.txt'
    );
  $out .= str_replace('<p>', '<' . $paragraph_start . '>', _inspect_readme2html($readme) )
      . '</div></div><hr/>';
  return $out;
}

/**
 * Parse README.txt to html.
 *
 *   Replacements:
 *   - formats headlines; line with no hyphens followed by all hyphens line
 *   - hyphen to bullet; newline+space+*+space becomes newline+space+bullet+space
 *   - emphasizes underscore fragments; _what ever_ becomes <em>what ever</em>
 *   - turns url into link, and http://your-drupal-site.tld/some-path becomes internal link to /some-path
 *   - turns mail address into mailto link
 *
 * Some replacements (like emphasizing) only works for ascii letters, not for letters line like ñ or ö.
 * @param string $readme_txt
 * @param string $headline_tag
 *   - default: h5
 * @return string
 */
function _inspect_readme2html($readme_txt, $headline_tag = 'h5') {
  if (!strlen($readme_txt)) {
    return 'empty readme';
  }
  $ndls = array(
    '/</',
    '/>/',
    '/\r\n|\r/', // CR -> NL
    '/[\x20\t]+\n/', // trailing spaces (in lines)
    '/\n{3,}/', // 3 or more newlines to double
    '/<([^>@]+@[^>@]+)>/', // mail addresses
    '/([\n\x20])_([a-zA-Z\d][^\.,;\:\n\x20]*[a-zA-Z\d])_([\n\x20\.,;\:])/', // _emphasize_ -> <em>emphasize</em>
  );
  $rplcs = array(
    '&lt;',
    '&gt;',
    "\n",
    "\n",
    "\n\n",
    '<a href="mailto:$1">&#60;$1&#62;</a>',
    '$1<em>$2</em>$3',
  );
  $s = preg_replace($ndls, $rplcs, $readme_txt);

  //  Insert links in CONTENTS OF THIS FILE.
  if (strpos($s, 'CONTENTS OF THIS FILE') !== FALSE) {
    $pos_start = strpos($s, ' * ');
    $pos_end = strpos($s, "\n\n", $pos_start);
    $s_toc = preg_replace('/[ ]+\n/', "\n", substr($s, $pos_start, ($pos_end - $pos_start) + 1) ); // remove trailing space(s)
    //echo $s_toc; exit;
    $s_toc = preg_replace_callback(
      '/ \* ([^\n]+)\n/',
      create_function(
          '$ms',
          '
          return \' * <a href="#\' .
              strtolower( preg_replace(\'/[^a-zA-Z_\-]/\', \'_\', $ms[1]) )
              . \'">\' . $ms[1] . \'</a>\' . "\n";
          '
      ),
      $s_toc
    );
    $s_start = substr($s, 0, $pos_start);
    $s_end = substr($s, $pos_end);
    $s = str_replace('CONTENTS OF THIS FILE', 'CONTENTS', $s_start) . $s_toc . $s_end;
  }

  //  Format headlines, and insert anchor.
  $s = preg_replace_callback(
      '/\n([^\n\-])([^\n]*[^\n\-])\n[\-]{2,}\n+/',
      create_function(
          '$ms',
          '
          return \'</p><a name="\'
              . strtolower( preg_replace(\'/[^a-zA-Z_\-]/\', \'_\', $ms[1] . $ms[2]) )
              . \'"></a><headlineTag style="margin-top:20px;">\'
              . $ms[1]
              . \'<span style="text-transform:lowercase;">\' . $ms[2] . \'</span></headlineTag><p>\';
          '
      ),
      $s
    );
  $s = str_replace('headlineTag', $headline_tag, $s);

  $ndls = array(
    '/^[^\n]*\n/', // remove first line
    //'//',
  );
  $rplcs = array(
    '',
    //'',
  );
  $s = preg_replace($ndls, $rplcs, $s);
  //  Links may be followed by a punctuation marker, hard to detect without a callback.
  //  And http://your-drupal-site.tld/path should become /path
  $s = preg_replace_callback(
      '/(https?\:\/\/)(\S+)([\x20\t\n])/',
      create_function(
          '$ms',
          '
          $protocol = $ms[1];
          $le = strlen($addr = $ms[2]);
          $dot = \'\';
          if (!preg_match(\'/[\/a-zA-Z\d]/\', $addr{$le - 1})) {
            $dot = $addr{$le - 1};
            $addr = substr($addr, 0, $le - 1);
          }
          if (strpos($addr, \'your-drupal-site.tld/\') === 0) {
            $protocol = \'/\';
            $addr = str_replace(\'your-drupal-site.tld/\', \'\', $addr);
          }
          return \'<a href="\' . $protocol . $addr . \'">\' . $addr . \'</a>\' . $dot . $ms[3];
          '
      ),
      $s
    );
  $ndls = array(
    '/([\n>]) \* /', // \n *  to \n bullet
    '/^\n?<\/p>/', // first ending <p> added by headline (in total)
    '/\n?$/', // trailing newlines (in total)
    '/\n{2,}/',
    //'//',
  );
  $rplcs = array(
    '$1 &bull; ',
    '',
    '',
    '</p><p>',
    //'',
  );
  return nl2br( preg_replace($ndls, $rplcs, $s) ) . '</p>';
}